/*
** isc.h                        structures and defines used in a ISC server
**
** Copyright (C) 1991, 1996, 1998-1999, 2001 by Peter Eriksson and
** Per Cederqvist of the Lysator Academic Computer Association.
**
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
** history:
** 910306 pen      major overhaul
** 910307 pen      type name changes, changes in structs..
** 920207 pen      updated
** (See ChangeLog for recent history)
*/

#ifndef __ISC_H__
#define __ISC_H__

/* FIXME (Bug 563): Don't include files here. */
#include <sys/socket.h>
#include <netinet/in.h>

/*
** Give the poor user a chance to change it
*/
#ifndef ISC_UDGTYPE
#  define ISC_UDGTYPE void
#endif

/* Forward declarations. */
struct isc_msgqueue;
struct isc_session_cfg;
struct isc_scb_entry;
struct isc_mcb;
struct isc_scb;

/*
 * Callback function types.
 */

/* When a new TCP client conntects.  */
typedef void *isc_accept_callback(struct isc_mcb *cb_mcb,
				  struct isc_scb *cb_accepting_session,
				  struct isc_scb *cb_new_session);

typedef void isc_write_error_cb(struct isc_mcb *cb_mcb,
				struct isc_scb *cb_session,
				int saved_errno,
				void *user);

/*
 * For IPv4/IPv6 compatibility:
 * SOCKADDR_STORAGE: Socket address storage type.
 * struct sockaddr_storage is used when available (RFC2553)
 */

/* FIXME (Bug 563): HAVE_STRUCT_SOCKADDR_STORAGE is only defined when
   compiling the isc library.  Move this to intern.h, if possible.  */
#ifdef HAVE_STRUCT_SOCKADDR_STORAGE
#define SOCKADDR_STORAGE struct sockaddr_storage
#else
#define SOCKADDR_STORAGE struct sockaddr
#endif


/*
** The different session states
*/
enum isc_session_state
{
  ISC_STATE_RUNNING,
  ISC_STATE_DISABLED,
  ISC_STATE_CLOSING,
  ISC_STATE_LISTENING,
};


/* Socket address
 */

union sockaddrs {
  SOCKADDR_STORAGE ssa;
  struct sockaddr sa;
  struct sockaddr_in sa_in;
  /* FIXME (Bug 563): USE_INET6 is only defined when compiling the
     isc library.  Move this to intern.h, if possible.  */
#ifdef USE_INET6
  struct sockaddr_in6 sa_in6;
#endif
};

/*
** Message information structure
*/
union isc_address
{
    union sockaddrs saddr;
};

/*
** Session control structure
*/
struct isc_scb
{
  int                 nlinks;
  
  enum isc_session_state state;
  int                 fd;
  
  struct isc_msgqueue *wr_msg_q;

  struct isc_session_cfg * cfg;
  union isc_address *raddr;
  union isc_address *laddr;
  
  /* Cheap transmit buffer */
  /* Should be dynamically allocated - into an struct isc_msg */
  char                sendbuf[2048];
  int                 sendindex;
  
  ISC_UDGTYPE       * udg;   /* User defined garbage :-) */

  struct isc_mcb      *master;
  
  isc_accept_callback *accept_cb;
  isc_write_error_cb  *write_err_cb;

  oop_call_fd         *data_available_callback;
  void                *data_available_user;
  int                  data_available_registered;

  int                  write_cb_registered;
};




/*
** The Master Control Block
*/
struct isc_mcb
{
  struct isc_session_cfg  * scfg;
  struct isc_scb_entry    * sessions;
  oop_source              * event_source;
};


/*
** Setup a set of functions to handle memory allocation
*/
extern void
isc_setallocfn(void  * (*mallocfn)(size_t size),
               void  * (*reallocfn)(void  * buf, size_t size),
	       void    (*freefn)(void  * buf));

/*
** This routine will setup the ISC subsystem
*/
extern struct isc_mcb *
isc_initialize(oop_source *);

extern void
isc_cfg_fd_relocate(struct isc_mcb *, int fd_relocate);


/*
** Shut down all sessions associated with an ISC Master Control Block
** and deallocate all storage used by the MCB.
*/
extern void
isc_shutdown(struct isc_mcb  * mcb);



/*
** Establish a TCP port to listen for connections at
*/

extern struct isc_scb *
isc_listentcp(struct isc_mcb   * mcb,
	      const char  * address,
	      const char  * service,
	      isc_accept_callback *cb);

/*
** The callback function supplied to isc_listentcp can call this
** function to install a callback that will be called when data is
** available on the new socket.
*/
extern void
isc_set_read_callback(struct isc_scb *session,
		      oop_call_fd *data_available_cb,
		      isc_write_error_cb *write_error_cb,
		      void *user);

extern int
isc_destroy(struct isc_mcb  * mcb, struct isc_scb  * scb);

/*
** Read data from a session into the end of a string.  The first
** UNUSED bytes of the string are unused, and may be reused.
*/
enum isc_read_result {
    ISC_READ_DATA,
    ISC_READ_LOGOUT,
    ISC_READ_WOULDBLOCK,
    ISC_READ_ERROR,		/* read() failed unexpectedly -- see errno. */
    ISC_READ_NOMEM,		/* out of memory. */
};

extern enum isc_read_result
isc_read_data(struct isc_scb *scb,
	      String *result,
	      String_size *unused);


 
/*
** Flush the transmit queue for a specific session.
** Unless this function is called, the output might remain in the
** output buffer forever.
*/
extern void
isc_flush(struct isc_scb  * scb);

/*
** Put a buffer on the transmit queue for a session
*/
extern int
isc_write(struct isc_scb *scb,
	  const void  * buffer,
	  size_t        length);



/*
** Put a single character on the transmit queue
*/
extern int
isc_putc(int           chr,
	 struct isc_scb *scb);

/*
** Put a NUL-terminated string on the transmit queue (the NUL is not sent).
*/
extern int
isc_puts(const char *str,
	 struct isc_scb *scb);

/*
** Put a decimal representation of ``nr'' on the transmit queue.
*/
extern int isc_putul(unsigned long nr, struct isc_scb *scb);

/*
** Enable a previously disabled session
*/
extern int
isc_enable(struct isc_scb *scb);

/*
** Disable a session. The session will not generate read events, but
** any enqueued output will still be sent to the remote host.
*/
extern int
isc_disable(struct isc_scb *scb);

extern union isc_address *
isc_getladdress(struct isc_scb *scb);

extern union isc_address *
isc_getraddress(struct isc_scb *scb);


extern void
isc_freeaddress(union isc_address *addr);

extern char *
isc_getipnum(union isc_address *ia,
	     char *address,
	     int len);

extern char *
isc_gethostname(union isc_address *ia,
		char *address,
		int len);


extern int
isc_getportnum(union isc_address *ia);

#endif /* __ISC_H__ */
