/*
** isc_session.c                              Routines to handle ISC sessions
**
** Copyright (C) 1991-1993, 1998-1999, 2001 by Peter Eriksson and
** Per Cederqvist of the Lysator Academic Computer Association.
**
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
** history:
** 910305 pen      moved into separate file
** 920129 pen      added support for "lazy" connect()
** 920209 pen      reworked some of the code
** 920209 pen      TCP and UDP specific code removed.
** 930117 pen      Two memory leak bugs fixed (found by Ceder)
** (See ChangeLog for recent history)
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <time.h>
#ifdef HAVE_STDDEF_H
#  include <stddef.h>
#endif
#ifdef HAVE_STDARG_H
#  include <stdarg.h>
#endif
#include <fcntl.h>
#ifndef NULL
#  include <stdio.h>
#endif
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#include <assert.h>

#include "oop.h"

#include "s-string.h"

#include "isc.h"
#include "intern.h"
#include "unused.h"


/*
** Insert a session into a master control structure
*/
int
isc_insert(struct isc_mcb *mcb,
	   struct isc_scb *scb)
{
  struct isc_scb_entry *isl;


  ISC_XNEW(isl);
  isl->scb = scb;

  /* Better check if this is the first session or not */
  if (mcb->sessions != NULL)
  {
    isl->next           = mcb->sessions;
    isl->prev           = mcb->sessions->prev;
    isl->prev->next     = isl;
    mcb->sessions->prev = isl;
  }
  else
  {
    isl->next = isl;
    isl->prev = isl;
  }
  
  mcb->sessions  = isl;

  return ++scb->nlinks;
}



/*
** Locate a session list entry in a MCB
*/
static struct isc_scb_entry *
isc_findsession(struct isc_mcb *mcb,
		struct isc_scb *scb)
{
  struct isc_scb_entry *isl;


  /* Locate the session in the MCB */
  for (isl = mcb->sessions; isl != NULL && isl->scb != scb; isl = isl->next)
    if (isl->next == mcb->sessions)
      break;

  /* Not found? Return -1 */
  if (isl == NULL || isl->scb != scb)
    return NULL;

  return isl;
}



/*
** Remove a session from a master control structure
*/
int
isc_remove(struct isc_mcb *mcb,
	   struct isc_scb *scb)
{
  struct isc_scb_entry *isl;


  isl = isc_findsession(mcb, scb);
  if (!isl)
  {
    errno = ENOENT;
    return -1;
  }
  
  /* Remove it from the list of sessions if it is still on it */
  if (isl->prev != NULL)
  {
    if (isl->next != isl)
    {
      /* Make sure noone references this session */
      if (mcb->sessions == isl)
	mcb->sessions = isl->next;

      isl->prev->next = isl->next;
      isl->next->prev = isl->prev;
    }
    else
    {
      /* This was the last session on the circular list. */
      mcb->sessions = NULL;
    }
    
    isl->prev = NULL;
    isl->next = NULL;
  }

  isc_free(isl);
  
  return --scb->nlinks;
}


/*
** Create a new session structure. 
*/
struct isc_scb *
isc_create(struct isc_mcb *mcb,
	   struct isc_session_cfg *cfg,
	   enum isc_session_state initial_state)
{
  struct isc_scb  * scb;
  
  
  ISC_XNEW(scb);

  scb->nlinks   = 0;
  scb->fd       = -1;
  scb->state    = initial_state;
  scb->wr_msg_q = NULL; 
  scb->accept_cb= NULL;
  scb->write_cb_registered = 0;
  scb->data_available_registered = 0;
  scb->data_available_callback = NULL;
  scb->data_available_user = NULL;

  scb->cfg = cfg;
  scb->master = mcb;
  
  /* Fill in the session structure */
  scb->wr_msg_q      = isc_newqueue();
  scb->sendindex     = 0;

  return scb;
}

void
isc_set_read_callback(struct isc_scb *session,
		      oop_call_fd *data_available_callback,
		      isc_write_error_cb *write_error_cb,
		      void *user)
{
  if (session->data_available_registered)
      isc_cancel_read_callback(session);
  session->data_available_callback = data_available_callback;
  session->data_available_user = user;
  session->write_err_cb = write_error_cb;
  isc_check_read_callback(session);
}

void
isc_check_read_callback(struct isc_scb *session)
{
  oop_source *source = session->master->event_source;

  int wanted = 1;

  /* Too many queued blocks?  Don't read any more data, until
     the client has read the output we have already produced. */
  if (isc_sizequeue(session->wr_msg_q) >= session->cfg->max.queuedsize)
    wanted = 0;

  if (session->state == ISC_STATE_DISABLED)
    wanted = 0;

  if (!wanted)
    isc_cancel_read_callback(session);
  else if (!session->data_available_registered
	   && session->data_available_callback != NULL)
  {
    source->on_fd(source, session->fd, OOP_READ,
		  session->data_available_callback,
		  session->data_available_user);
    session->data_available_registered = 1;
  }
}

void
isc_cancel_read_callback(struct isc_scb *session)
{
  oop_source *source = session->master->event_source;

  if (session->data_available_registered)
  {
    source->cancel_fd(source, session->fd, OOP_READ);
    session->data_available_registered = 0;
  }
}

void
isc_cancel_write_callback(struct isc_scb *session)
{
  oop_source *source = session->master->event_source;
      
  if (session->write_cb_registered)
  {
    source->cancel_fd(source, session->fd, OOP_WRITE);
    session->write_cb_registered = 0;
  }
}


/*
** Destroy the allocated Session structure. Close the
** socket if open and if Master structure specified, remove
** the session from it.
*/
int
isc_destroy(struct isc_mcb *mcb,
	    struct isc_scb *scb)
{
  int code = 0;

  isc_cancel_read_callback(scb);
  isc_cancel_write_callback(scb);

  if (mcb)
  {
    code = isc_remove(mcb, scb);
    if (code < 0)
      return code;
    if (scb->accept_cb != NULL)
    {
      mcb->event_source->cancel_fd(mcb->event_source, scb->fd, OOP_READ);
      scb->accept_cb = NULL;
    }
  }

  assert(scb->accept_cb == NULL);

  /*
  ** Only close and deallocate storage if last reference has been
  ** removed. 
  */
  if (scb->nlinks <= 0)
  {
    if (scb->raddr)
    {
      isc_freeaddress(scb->raddr);
      scb->raddr = NULL;
    }
  
    if (scb->laddr)
    {
      isc_freeaddress(scb->laddr);
      scb->laddr = NULL;
    }
    
    if (scb->fd != -1)
    {
      close(scb->fd);
      scb->fd = -1;
    }

    if (scb->wr_msg_q)
      isc_killqueue(scb->wr_msg_q);

    isc_free(scb);
  }

  return code;
}

int isc_disable(struct isc_scb *scb)
{
  if (scb->state != ISC_STATE_RUNNING)
    return -1;
  
  scb->state = ISC_STATE_DISABLED;
  isc_check_read_callback(scb);
  return 0;
}



int isc_enable(struct isc_scb *scb)
{
  if (scb->state != ISC_STATE_DISABLED)
    return -1;
  
  scb->state = ISC_STATE_RUNNING;
  isc_check_read_callback(scb);
  return 0;
}
